import numpy as np
import matplotlib as mpl
from matplotlib.backends import backend_svg
import matplotlib.pyplot as plt
import seaborn as sns

import io
import numbers


def print_bbox_from_arr(bbox):
    print(f'Bbox(xmin={bbox[0, 0]}, xmax={bbox[1, 0]},\n'
          f'     ymin={bbox[0, 1]}, ymax={bbox[1, 1]}')


def print_extents_from_arr(bbox):
    print(f'Bbox(xmin={bbox[0, 0]}, ymin={bbox[0, 1]},\n'
          f'     width={bbox[1, 0] - bbox[0, 0]}, '
          f'height={bbox[1, 1] - bbox[0, 1]})')


def plot_colored_line(t, x, y, cmap='viridis', linewidth=3, ax=None,
                      colorbar=True):
    points = np.array([x, y]).T.reshape(-1, 1, 2)
    segments = np.concatenate([points[:-1], points[1:]], axis=1)

    lc = mpl.collections.LineCollection(segments, cmap=plt.get_cmap(cmap))
    lc.set_array(t)  # Collection is a ScalarMappable
    lc.set_linewidth(linewidth)

    if ax is None:
        fig, ax = plt.subplots()
    ax.add_collection(lc)
    ax.set_xlim(get_lim(x))
    ax.set_ylim(get_lim(y))

    if colorbar:
        cnorm = mpl.colors.Normalize(vmin=np.min(t), vmax=np.max(t))
        sm = plt.cm.ScalarMappable(cmap=cmap, norm=cnorm)
        sm.set_array(t)
        # can't find a way to set a colorbar simply without grabbing the
        # current axis, so make sure we can restore what the "current axis"
        # was before we did this
        cax = plt.gca()
        plt.sca(ax)
        plt.colorbar(sm)
        plt.sca(cax)
    return ax


def get_lim(x, margin=0.1):
    min = np.min(x)
    max = np.max(x)
    dx = max - min
    return [min - margin*dx, max + margin*dx]


def cmap_from_list(labels, palette=None, log=False, vmin=None, vmax=None):
    # sequential colormap if numbers
    if isinstance(labels[0], numbers.Number):
        labels = np.array(labels)
        if vmin is None:
            vmin = labels.min()
        if vmax is None:
            vmax = labels.max()
        if log:
            cnorm = mpl.colors.LogNorm(vmin=vmin, vmax=vmax)
        else:
            cnorm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)
        if palette is None:
            palette = 'viridis'
        cmap = mpl.cm.get_cmap(palette)
        return lambda l: cmap(cnorm(l))
    # otherwise categorical map
    else:
        if log:
            raise ValueError('LogNorm makes no sense for categorical labels.')
        labels = list(set(labels))
        n_labels = len(labels)
        pal = sns.color_palette(palette, n_colors=n_labels)
        cmap = {labels[i]: pal[i] for i in range(n_labels)}
        return lambda l: cmap[l]


def draw_triangle(alpha, x0, width, orientation, base=10, **kwargs):
    """Draw a triangle showing the best-fit slope on a linear scale.

    Parameters
    ----------
    alpha : float
        the slope being demonstrated
    x0 : (2,) array_like
        the "left tip" of the triangle, where the hypotenuse starts
    width : float
        horizontal size
    orientation : string
        'up' or 'down', control which way the triangle's right angle "points"
    base : float
        scale "width" for non-base 10

    Returns
    -------
    corner : (2,) np.array
        coordinates of the right-angled corner of the triangle
    """
    x0, y0 = x0
    x1 = x0 + width
    y1 = y0 + alpha*(x1 - x0)
    plt.plot([x0, x1], [y0, y1], 'k')
    if (alpha >= 0 and orientation == 'up') \
            or (alpha < 0 and orientation == 'down'):
        plt.plot([x0, x1], [y1, y1], 'k')
        plt.plot([x0, x0], [y0, y1], 'k')
        # plt.plot lines have nice rounded caps
        # plt.hlines(y1, x0, x1, **kwargs)
        # plt.vlines(x0, y0, y1, **kwargs)
        corner = [x0, y1]
    elif (alpha >= 0 and orientation == 'down') \
            or (alpha < 0 and orientation == 'up'):
        plt.plot([x0, x1], [y0, y0], 'k')
        plt.plot([x1, x1], [y0, y1], 'k')
        # plt.hlines(y0, x0, x1, **kwargs)
        # plt.vlines(x1, y0, y1, **kwargs)
        corner = [x1, y0]
    else:
        raise ValueError(r"Need $\alpha\in\mathbb{R} and "
                         r"orientation\in{'up', 'down'}")
    return corner


def draw_power_law_triangle(alpha, x0, width, orientation, base=10,
                            x0_logscale=True, label=None,
                            label_padding=0.1, text_args={}, ax=None,
                            **kwargs):
    """Draw a triangle showing the best-fit power-law on a log-log scale.

    Parameters
    ----------
    alpha : float
        the power-law slope being demonstrated
    x0 : (2,) array_like
        the "left tip" of the power law triangle, where the hypotenuse starts
        (in log units, to be consistent with draw_triangle)
    width : float
        horizontal size in number of major log ticks (default base-10)
    orientation : string
        'up' or 'down', control which way the triangle's right angle "points"
    base : float
        scale "width" for non-base 10
    ax : mpl.axes.Axes, optional

    Returns
    -------
    corner : (2,) np.array
        coordinates of the right-angled corhow to get text outline of the
        triangle

    """
    if x0_logscale:
        x0, y0 = [base**x for x in x0]
    else:
        x0, y0 = x0
    if ax is None:
        ax = plt.gca()
    x1 = x0*base**width
    y1 = y0*(x1/x0)**alpha
    ax.plot([x0, x1], [y0, y1], 'k')
    if (alpha >= 0 and orientation == 'up') \
            or (alpha < 0 and orientation == 'down'):
        ax.plot([x0, x1], [y1, y1], 'k')
        ax.plot([x0, x0], [y0, y1], 'k')
        # plt.plot lines have nice rounded caps
        # plt.hlines(y1, x0, x1, **kwargs)
        # plt.vlines(x0, y0, y1, **kwargs)
        corner = [x0, y1]
    elif (alpha >= 0 and orientation == 'down') \
            or (alpha < 0 and orientation == 'up'):
        ax.plot([x0, x1], [y0, y0], 'k')
        ax.plot([x1, x1], [y0, y1], 'k')
        # plt.hlines(y0, x0, x1, **kwargs)
        # plt.vlines(x1, y0, y1, **kwargs)
        corner = [x1, y0]
    else:
        raise ValueError(r"Need $\alpha\in\mathbb{R} and orientation\in{'up', "
                         r"'down'}")
    if label is not None:
        xlabel = x0*base**(width/2)
        if orientation == 'up':
            ylabel = y1*base**label_padding
        else:
            ylabel = y0*base**(-label_padding)
        ax.text(xlabel, ylabel, label, horizontalalignment='center',
                verticalalignment='center', **text_args)
    return corner


def set_ax_size(w, h, ax=None):
    """pass absolute height of axes, not figure

    gets around fact that throughout matplotlib, only full figure aspect ratio
    is under user control.

    w, h: width, height in inches """
    if not ax:
        ax = plt.gca()
    l = ax.figure.subplotpars.left  # NOQA
    r = ax.figure.subplotpars.right
    t = ax.figure.subplotpars.top
    b = ax.figure.subplotpars.bottom
    figw = float(w)/(r-l)
    figh = float(h)/(t-b)
    ax.figure.set_size_inches(figw, figh)


# there's quite a fight over whether ndarray should be a
# "collections.abc.Sequence" (https://github.com/numpy/numpy/issues/2776), so
# instead we check the attrs we want ourselves
def _has_len_get_item(x):
    return hasattr(x, '__len__') and hasattr(x, '__getitem__')


def _ensure_seq(arg, N, unit_type=None, name=None, match_name=None):
    """
    Encapsulate the logic for tiling an input parameter.

    Parameters
    ----------
    arg : unit_type or list thereof
        The input argument you want to ensure is a list.
    N : int
        The length the output list should match
    unit_type : type
        The type of the input argument (if it's not a list, or the type of each
        element of the output list).
    name : str, optional
        The name of the argument, for error messages. If ``None``, the argument
        will be str'd.
    match_name : str, optional
        The name of the parameter whose length *arg* is intended to match. If
        ``None``, the desired length itself will be printed.

    Returns
    -------
    list of unit_type
        The input parameter, now guaranteed to be a list of length *N*.

    """
    if name is None:
        name = arg
    if match_name is None:
        match_name = f' be {N}.'
    else:
        match_name = ' match {match_name}.'

    # a nested unit_type
    nested_sequence = (
        unit_type is not None
        and isinstance(arg, unit_type)
        and _has_len_get_item(unit_type)
        and isinstance(arg[0], unit_type)
    )
    # the unit type is itself a sequence, but not nested
    unit_is_seq = (
        unit_type is not None
        and isinstance(arg, unit_type)
        and _has_len_get_item(unit_type)
        and not isinstance(arg[0], unit_type)
    )
    # if the input is already tiled, verify its length
    if nested_sequence or (not unit_is_seq and _has_len_get_item(arg)):
        if N != len(arg):
            raise ValueError(f"length of {name} should {match_name}.")
    # otherwise, if we've been given a type, make sure it matches
    elif unit_type is not None and not isinstance(arg, unit_type):
        raise ValueError(f"{name} should be a {unit_type} or list thereof.")
    # otherwise, we have a single instance of the desired input, tile it
    else:
        arg = N*[arg]
    return np.array(arg)


def make_at_aspect(plot_funcs, heights, col_width,  tight_width='bbox',
                   hspace=None, halign='min_axis_width', is_ratio=True,
                   **kw_figs):
    """Figure with fixed column width and requested aspect ratios per subplot.

    Parameters
    ----------
    plot_funcs : Sequence[Callable[[matplotlib.axes.Axes], None]]
        function to create each subplot given the axes to draw it into
    heights : float or Sequence[float]
        If is_ratio is True, this is the height/width ratio desired for each
        subplot. A single number is allowed for fixed aspect ratio across all
        axes.  If is_ratio is False, this is the absolute height desired.
    col_width : float
        absolute width of figure desired, in inches
    tight_width : string
        'bbox' or 'tight'. 'bbox' means use default generated by
        bbox_inches='tight'. 'tight' means to actually remove all whitespace
        from left and right size of final figure. This option is more a
        reflection of my lack of interest in figuring out which of these is the
        "correct" thing to do as opposed to actually useful.
    hspace : float
        vertical space between adjacent axes bboxes as fraction of average axis
        height (not fraction of average axis bbox height). defaults to value in
        rc (fig.subplotpars.hspace).
    halign : {'min_axis_width', 'full'} or Sequence[string]
        'full' means make each axis as wide as it can be while still fitting in
        the figure boundary. 'min_axis_width' means make the final axis width
        match the width of the smallest axis (so that multiple axes with
        different amounts of required label padding still line up).
    is_ratio : bool or Sequence[bool]
        Whether each height request is a ratio or absolute axis height desired.


    Returns
    -------
    ax : matplotlib.figure.Figure
        the final figure

    Notes
    -----
    No extra "axes_kw" argument is provided, since each axis is passed to it's
    appropriate plot_func, which should be able to set any relevant extra
    parameters for that axis.

    """
    n_plots = len(plot_funcs)
    halign = _ensure_seq(halign, n_plots, str, 'halign', 'plot_funcs')
    heights = _ensure_seq(heights, n_plots, numbers.Number,
                          'heights', 'plot_funcs')
    is_ratio = _ensure_seq(is_ratio, n_plots, bool, 'is_ratio', 'plot_funcs')
    # first make "test" figure to get correct extents including all labels, etc
    # leave a 2x margin of error for the labels, etc. to fit into vertically.
    max_heights = heights.copy()
    max_heights[is_ratio] = heights[is_ratio]*col_width
    test_fig_height = 2*np.sum(max_heights)
    fig, axs = plt.subplots(nrows=n_plots,
                            figsize=(col_width, test_fig_height),
                            **kw_figs)
    if n_plots == 1:
        axs = [axs]  # not sure why the inconsistency in subplots interface...
    for i in range(n_plots):
        plot_funcs[i](axs[i])

    # "dry-run" a figure save to get "real" bbox for full figure with all
    # children
    fake_file = io.StringIO()
    fig.canvas = backend_svg.FigureCanvasSVG(fig)
    renderer = backend_svg.RendererSVG(col_width, test_fig_height, fake_file)
    _ = fig.draw(renderer)

    disp_to_inch = fig.dpi_scale_trans.inverted()
    fig_to_inch = fig.dpi_scale_trans.inverted() + fig.transFigure

    # now get extents of axes themselves and axis "bbox"s (i.e. including
    # labels, etc)
    ax_bbox_inches = []
    ax_inches = []
    for ax in axs:
        # get_tightbbox docs say in "figure pixels", but means "display"
        ax_bbox = ax.get_tightbbox(renderer)
        ax_bbox_inches.append(disp_to_inch.transform(ax_bbox))
        ax_inches.append(fig_to_inch.transform(ax.get_position()))
    x0s, y0s, x1s, y1s = map(np.array, zip(
        *map(np.ndarray.flatten, ax_bbox_inches)
    ))
    ax0s, ay0s, ax1s, ay1s = map(np.array, zip(
        *map(np.ndarray.flatten, ax_inches)
    ))

    # calculate space required in addition to axes themselves
    pads_left = ax0s - x0s
    pads_right = x1s - ax1s
    pads_above = y1s - ay1s
    pads_below = ay0s - y0s
    max_pad_left = np.max(pads_left)
    max_pad_right = np.max(pads_right)

    # get axes sizes
    x0 = np.zeros(n_plots)
    x1 = np.zeros(n_plots)
    for i in range(n_plots):
        if halign[i] == 'full':
            x0[i] = pads_left[i]
            x1[i] = col_width - pads_right[i]
        elif halign[i] == 'min_axis_width':
            x0[i] = max_pad_left
            x1[i] = col_width - max_pad_right
        else:
            raise ValueError(f"Invalid option passed for halign: {halign[i]}. "
                             f"Should be 'full' or 'min_axis_width'")
    real_heights = heights.copy()
    real_heights[is_ratio] = heights[is_ratio]*(x1 - x0)[is_ratio]
    if hspace is None:
        hspace = fig.subplotpars.hspace
    hspace = hspace*np.mean(real_heights)

    # make new figure with axes "correctly" located
    total_height = np.sum(real_heights) + np.sum(pads_below) \
        + np.sum(pads_above) + (n_plots - 1)*hspace
    fig = plt.figure(figsize=(col_width, total_height), **kw_figs)
    cur_y = 1  # track y pos normalized to height of figure
    for i in range(n_plots):
        left = x0[i]/col_width
        right = x1[i]/col_width
        cur_y -= pads_above[i]/total_height
        top = cur_y
        cur_y -= real_heights[i]/total_height
        bottom = cur_y
        cur_y -= pads_below[i]/total_height
        cur_y -= hspace/total_height
        if right < left or top < bottom:
            raise RuntimeError(f"Annoations requested for plot in {i}th Axes "
                               "take up so much space that there's no room "
                               "left for the Axes!")
        ax = fig.add_axes([left, bottom, right - left, top - bottom])
        plot_funcs[i](ax)
    return fig, real_heights



import numpy as np
from scipy.special import erfc
from .mittag_leffler import ml

def test_ml_exp():
    z = np.linspace(-2., 2., 50)
    assert np.allclose(ml(z, 1.), np.exp(z))

def test_ml_cosh():
    z = np.linspace(-2., 2., 50)
    assert np.allclose(ml(z**2, 2.), np.cosh(z))

def test_ml_experfc():
    z = np.linspace(0., 2., 50)
    assert np.allclose(ml(np.sqrt(z), 0.5), np.exp(z)*erfc(-np.sqrt(z)))



import numpy as np
from numba import jit

from functools import reduce

@jit(nopython=True)
def where(element, test_elements, assume_sorted=True):
    """Return indices of `test_element` containing each element of
    `element` (np.nan otherwise), for case where `test_elements` is an array of unique
    values. Broadcasts over `element`, so returned array has that same size.

    Parameters
    ----------
    element : (N,) array_like
        elements to search for in `test_elements`
    test_elements : (M,) array_like
        array to extract locations out of
    assume_sorted : bool
        whether or not test_elements is sorted. will internally sort if False

    Returns
    """
    if assume_sorted:
        ixs = np.arange(len(test_elements))
    else:
        ixs = np.argsort(test_elements)
        test_elements = test_elements[ixs]
    elem_ix = np.argsort(element)
    elem = element[elem_ix]
    wi = np.zeros_like(element)
    ti = 0
    for ei, e in enumerate(elem):
        #TODO: if log(len(unique(elem))) <<< elem, use binary search instead of
        # linear search here for big speedup
        while test_elements[ti] < e:
            ti += 1
        if test_elements[ti] > e:
            wi[ei] = np.nan
        else:
            wi[ei] = ixs[ti] # works whether or not assume_sorted
    return wi

def loglinspace(Ntot, Nlog, Nlin):
    """Union of linspaces with log-spaced ends, same start.

    Whenever you want to use simulation output to calculate some type of
    correlation (e.g. MSD from BD), you typically want to linearly space points
    so that you can use sliding windows to easily compute the time-averaged
    correlation function.

    But then, in order to get statistics for the whole range of time scales
    covered by the simulation, it would naively seem that one needs to save
    *every* time point, which is extremely wasteful.

    Instead, one can save linear spaced points with different frequencies, to
    cover the different orders of magnitude. For example, you might simulate
    using the grid:

        >>> import numpy as np
        >>> t = np.arange(0, 1e8, 1e-2)

    but only save at times:

        >>> from functools import reduce
        >>> save_time_arrs = [np.arange(0, 1e2, 1e-2),
                np.arange(0, 1e4, 1e0), np.arange(0, 1e6, 1e2),
                np.arange(0, 1e8, 1e4)]
        >>> t_save = reduce(np.union1d, save_time_arrs)

    if we have the simulation output at any of the individual members of
    `save_time_arrs`, we can use a simple sliding window to calculate
    time-averaged correlation functions of our system over (in this case)
    `1e4` time points.

    But if we pass just `t_save` to a simulation function, the output will have
    each of these different evenly-space arrays interleaved with each other.
    The purpose of this function is to return an appropriate `t_save` given `t`
    and a specification of how much overlap there should be between
    adjacent `linspace`'s, as well as an array of indices for extracting each
    `save_time_arrs` member from `t_save` (and thus, how to extract
    evenly-spaced simulation time points from the interleaved, quasi-log-spaced
    simulation time points requested by `t_save`.

    Parameters
    ----------
    Ntot : int
        length of `t`
    Nlog : int
        number of log-spaced intervals
    Nlin : int
        number of linearly-space points per interval

    Returns
    -------
    t_save : (<= Nlog*Nlin,) array_like
        subset of t
    save_i : List<np.array>
        a Nlog-length list of Nlin-long 1d arrays of indices. for each `i`,
        t_save[save_i[i]] will extract one of the linearly-space sets of points

    Usage
    -----
    In order to  generate the example from earlier in this docstring, simply
    use:

        >>> t = np.linspace(0, 1e8, 1e10+1)
        >>> t_save, save_i = loglinspace(t, 4, 1e4)

    Then you can do, for example, some BD simulations:

        >>> import wlcsim.bd as bd
        >>> x = bd.rouse.jit_rouse(N=101, L=100, b=1, D=1, t=t, t_save=t_save)

    and compute the taMSD from this trajectory over all 10 orders of magnitude
    in time using "blocks" of `1e4` points at a time like:

        >>> msds = np.zeros_like(t_save)
        >>> for si in save_i:
        >>>     ts = t_save[si]
        >>>     c, m = bd.runge_kutta._get_bead_msd(x[si], 51) # mid-point bead
        >>>     # recombine into single array, overwriting values computed from
        >>>     # more tightly-space points with those from less tightly-spaced
        >>>     msds[si] = m/c
        >>>     # or just plot directly
        >>>     plt.plot(ts[1:], m[1:]/c[1:], 'k.')

    """
    raise NotImplementedError("Doesn't do as advertised....docstring is out of sync.")
    end = np.log10(Ntot)
    log_step = end/Nlog
    # e.g. np.log10(1e8)/4 -> step=2, ends=(2, 4, 6, 8)
    log_ends = np.arange(log_step, end+1, log_step)
    save_time_arrs = [np.linspace(0, np.power(10, e), Nlin) for e in log_ends]
    t_save = reduce(np.union1d, save_time_arrs)
    save_i = [where(t, t_save) for t in save_time_arrs]
    return t_save, save_i

def loglinrange(start, stop, dt, Nlin, overlap=0):
    """As linlogspace but an easier API.

    Chooses a subset of `np.arange(start, stop, dt)` that is made of linearly
    spaced points of differing distances apart, so that you only need
    `Nlin*np.log((stop-start)/dt)` points to cover the interval.

    Parameters
    ----------
    start : float
        The start of the first linspace'd region
    stop : float
        The end of the last linspace'd region
    dt : float
        The step size for the smallest region, i.e. `np.arange(start, start+(Nlin+1)*dt, dt)`
    Nlin : float
        the size of the linspace'd regions (in # points)
    overlap : float, :math:`\in[0,1)`
        how much the next region should overlap with the previous in log space
    """
    if overlap < 0 or overlap >= 1:
        raise ValueError("`overlap` must be in [0,1)")
    # index of each t_arr at which next t_arr should start to create the
    # correct `overlap` (log spaced interpolation)
    # at 1-overlap=0, we want the multiplier to be 1 and the index to be 0
    # (start over at same place)
    # at 1-overlap=1, we want the multiplier to be `Nlin`, and the index will
    # not exist (Nlin+1)
    #TODO: BUG: this works at overlap=0 just fine, but otherwise there seems to
    # be an offset between loglinrange and loglinsample, which grows larger as
    # overlap increases
    log_i = np.interp(x=(1 - overlap), xp=[0, 1], fp=[np.log10(1), np.log10(Nlin)])
    overlap_i = np.round(np.power(10.0, log_i)).astype(int)

    t = start
    t_arrs = []
    while t < stop:
        t_end = t + (Nlin+1)*dt
        t_arr = np.arange(t, t_end, dt)
        t_arrs.append(t_arr[t_arr < stop])
        # still on np.arange(start, stop, dt) grid by selecting directly from it
        t = t_arr[overlap_i]
        # since overlap_i is rounded, should stay on original grid as well
        dt *= overlap_i
    t_save = reduce(np.union1d, t_arrs)
    save_i = [where(t, t_save) for t in t_arrs]
    return t_save, save_i

def test_loglinrange():
    """Should remain a subset of np.arange when applicable"""
    t, i = bnp.loglinrange(0, 1e4, 1, 1e2, 0.5)
    t_all = np.arange(0, 1e4, 1)
    assert(np.all(np.isin(t, t_all)))
    # # this fails because of floating round-off
    # t, i = bnp.loglinrange(0, 1e4, 0.17, 1e2, 0.5)
    # t_all = np.arange(0, 1e4, 0.17)
    # assert(np.all(np.isin(t, t_all)))

def loglinsample(Ntot, Nlin, overlap):
    """Preferred version of loglin* functions.

    Sometimes, because of rounding errors,

        >>> np.isin(bnp.loglinrange(start, stop, dt, ...),
                    np.arange(start, stop, dt))

    will have some `False` values, because the two arrays are off by floating
    point mismatch sized differences.

    This function instead of returning absolute times, returns indices into the
    original `np.arange(start, stop, dt)` array, so that we can ensure no such
    error occurs."""
    t, i = loglinrange(0, Ntot+1, 1, Nlin, overlap)
    return np.round(t).astype(int), i

def test_loglinsample():
    """compare to loglinrange.

    WARNING: BUG: apparently only works currently for small enough `overlap`"""
    istart = 0
    stop = 1e4
    dt = 0.17
    Nlin = 1e2
    overlap = 0.2 # for overlap >= 0.3ish, this test fails

    # actual baseline grid
    t_all = np.arange(start, stop, dt)
    Ntot = len(t_all)
    # get spacing via sub-sampling
    ti, ii = bnp.loglinsample(Ntot, Nlin, overlap)
    ts = t_all[ti]
    # get spacing via direct sampling
    t, i = bnp.loglinrange(start, stop, dt, Nlin, overlap)
    # compare results
    assert(np.all(np.isclose(t, ts)))



import os
import struct
import numpy as np
# from .utils import well_behaved_decorator
from functools import wraps

def strong_default_seed(f):
    """Use os.random to correctly seed a np.random.RandomState-compatible
    random variate if the user does not provide the "seed" random variable. The
    correctly initialized state is passed to the function via the "random_state"
    kwarg.

    Parameters
    ----------
    f : Callable
        function that takes at least two kwargs, seed and random_state. random_state should
        be used to generate random numbers within the function.

    Returns
    -------
    wrapped_f : Callable
        Function that initializes random_state (using seed, or by generating a seed if
        none is given) if it is not passed to the function.

    Notes
    -----
    On sane operating systems, you should be able to parallelize the function
    and get different results for each invocation by letting this wrapper
    initiazlie the seed at each call site.
    """
    @wraps(f)
    def wrapped_f(*args, **kwargs):
        if 'random_state' in kwargs and kwargs['random_state'] is not None:
            return f(*args, **kwargs)
        if 'seed' not in kwargs or kwargs['seed'] is None:
            seed = struct.unpack("<L", os.urandom(4))[0]
        else:
            seed = kwargs['seed']
        random_state = np.random.RandomState()
        random_state.seed(seed)
        kwargs['random_state'] = seed
        return f(*args, **kwargs)
    return wrapped_f

def make_pool(rvs_f, pool_size=1024, unit_size=tuple(), **kwargs):
    """Use a random variate generator that takes a "size" keyword to make a
    function that just return random values "one" at a time from a pre-computed
    pool, replenishing the pool as necessary.

    This often provides serious speedups over calling the function with
    size=unit_size repeatedly when usign numpy/scipy random generation
    functions.

    .. warning::

        `kwargs` are passed on to the `rvs_f` function when it is called to
        replenish the pool. This means if `kwargs` is constantly changing, many
        of its values will be ignored.

    Parameters
    ----------

    rvs_f : Callable
        function that takes size kwarg and generates that many random numbers
    pool_size : int
        how many sets of random numbers to generate ahead of time. this is how
        many times the function we generate can be called before replenishing
        its pool.
    unit_size : Tuple[int]
        the shape of the set of random numbers to be returned every time the
        wrapped function is called

    Returns
    -------

    wrapped_rvs_f : Callable
        function that returns random array of shape unit_size every time it is
        called (ignores all arguments) from a precomputed pool.
    """
    def rv():
        if rv.ri == pool_size:
            rv.rand_pool = rvs_f(size=((pool_size, )+unit_size), **kwargs)
            rv.ri = 0
        rv.ri += 1
        return rv.rand_pool[rv.ri - 1]
    rv.rand_pool = rvs_f(size=((pool_size,)+unit_size), **kwargs)
    rv.ri = 0
    return rv




import numpy as np
from numba import jit

# for testing
import pandas as pd
import scipy.stats
import matplotlib.pyplot as plt

def rkf( f, a, b, x0, tol, hmax, hmin ):
    """Runge-Kutta-Fehlberg method to solve x' = f(x,t) with x(t[0]) = x0.

    USAGE:
        t, x = rkf(f, a, b, x0, tol, hmax, hmin)

    INPUT:
        f     - function equal to dx/dt = f(x,t)
        a     - left-hand endpoint of interval (initial condition is here)
        b     - right-hand endpoint of interval
        x0    - initial x value: x0 = x(a)
        tol   - maximum value of local truncation error estimate
        hmax  - maximum step size
        hmin  - minimum step size

    OUTPUT:
        t     - NumPy array of independent variable values
        x     - NumPy array of corresponding solution function values

    NOTES:
        This function implements 4th-5th order Runge-Kutta-Fehlberg Method
        to solve the initial value problem

           dx
           -- = f(x,t),     x(a) = x0
           dt

        on the interval [a,b].

        Based on pseudocode presented in "Numerical Analysis", 6th Edition,
        by Burden and Faires, Brooks-Cole, 1997.
    """

    # Coefficients used to compute the independent variable argument of f

    a2  =   2.500000000000000e-01  #  1/4
    a3  =   3.750000000000000e-01  #  3/8
    a4  =   9.230769230769231e-01  #  12/13
    a5  =   1.000000000000000e+00  #  1
    a6  =   5.000000000000000e-01  #  1/2

    # Coefficients used to compute the dependent variable argument of f

    b21 =   2.500000000000000e-01  #  1/4
    b31 =   9.375000000000000e-02  #  3/32
    b32 =   2.812500000000000e-01  #  9/32
    b41 =   8.793809740555303e-01  #  1932/2197
    b42 =  -3.277196176604461e+00  # -7200/2197
    b43 =   3.320892125625853e+00  #  7296/2197
    b51 =   2.032407407407407e+00  #  439/216
    b52 =  -8.000000000000000e+00  # -8
    b53 =   7.173489278752436e+00  #  3680/513
    b54 =  -2.058966861598441e-01  # -845/4104
    b61 =  -2.962962962962963e-01  # -8/27
    b62 =   2.000000000000000e+00  #  2
    b63 =  -1.381676413255361e+00  # -3544/2565
    b64 =   4.529727095516569e-01  #  1859/4104
    b65 =  -2.750000000000000e-01  # -11/40

    # Coefficients used to compute local truncation error estimate.  These
    # come from subtracting a 4th order RK estimate from a 5th order RK
    # estimate.

    r1  =   2.777777777777778e-03  #  1/360
    r3  =  -2.994152046783626e-02  # -128/4275
    r4  =  -2.919989367357789e-02  # -2197/75240
    r5  =   2.000000000000000e-02  #  1/50
    r6  =   3.636363636363636e-02  #  2/55

    # Coefficients used to compute 4th order RK estimate

    c1  =   1.157407407407407e-01  #  25/216
    c3  =   5.489278752436647e-01  #  1408/2565
    c4  =   5.353313840155945e-01  #  2197/4104
    c5  =  -2.000000000000000e-01  # -1/5

    # Set t and x according to initial condition and assume that h starts
    # with a value that is as large as possible.

    t = a
    x = x0
    h = hmax

    # Initialize arrays that will be returned

    T = np.array( [t] )
    X = np.array( [x] )

    while t < b:

        # Adjust step size when we get to last interval

        if t + h > b:
            h = b - t;

        # Compute values needed to compute truncation error estimate and
        # the 4th order RK estimate.

        k1 = h * f( x, t )
        k2 = h * f( x + b21 * k1, t + a2 * h )
        k3 = h * f( x + b31 * k1 + b32 * k2, t + a3 * h )
        k4 = h * f( x + b41 * k1 + b42 * k2 + b43 * k3, t + a4 * h )
        k5 = h * f( x + b51 * k1 + b52 * k2 + b53 * k3 + b54 * k4, t + a5 * h )
        k6 = h * f( x + b61 * k1 + b62 * k2 + b63 * k3 + b64 * k4 + b65 * k5, \
                    t + a6 * h )

        # Compute the estimate of the local truncation error.  If it's small
        # enough then we accept this step and save the 4th order estimate.

        r = abs( r1 * k1 + r3 * k3 + r4 * k4 + r5 * k5 + r6 * k6 ) / h
        if len( np.shape( r ) ) > 0:
            r = np.max( r )
        if r <= tol:
            t = t + h
            x = x + c1 * k1 + c3 * k3 + c4 * k4 + c5 * k5
            T = np.append( T, t )
            X = np.append( X, [x], 0 )

        # Now compute next step size, and make sure that it is not too big or
        # too small.
        # Wikipedia (on page for "adaptive stepsize", uses richardson
        # extrapolation also, but chooses
        # h*min(max(0.9*(tol/r, 0.3), 2)
        h = h * min( max( 0.84 * ( tol / r )**0.25, 0.1 ), 4.0 )

        if h > hmax:
            h = hmax
        elif h < hmin:
            print( "Error: stepsize should be smaller than %e." % hmin )
            break

    # endwhile

    return ( T, X )


def rkfm(f, D, a, b, x0, tol, hmax, hmin):
    """Runge-Kutta-Fehlberg-Maruyama method to solve x' = f(x,t) + R with
    x(t[0]) = x0, where R is an uncorrelated Gaussian process with constant
    diffusivity D.

    USAGE:
        t, x = rkf(f, D, a, b, x0, tol, hmax, hmin)

    INPUT:
        f     - function equal to dx/dt = f(x,t)
        a     - left-hand endpoint of interval (initial condition is here)
        b     - right-hand endpoint of interval
        x0    - initial x value: x0 = x(a)
        tol   - maximum value of local truncation error estimate
        hmax  - maximum step size
        hmin  - minimum step size

    OUTPUT:
        t     - NumPy array of independent variable values
        x     - NumPy array of corresponding solution function values

    NOTES:
        This function implements 4th-5th order Runge-Kutta-Fehlberg Method
        to solve the initial value problem

           dx
           -- = f(x,t),     x(a) = x0
           dt

        on the interval [a,b].

        Based on pseudocode presented in "Numerical Analysis", 6th Edition,
        by Burden and Faires, Brooks-Cole, 1997.
    """

    # Coefficients used to compute the independent variable argument of f
    a2  =   2.500000000000000e-01  #  1/4
    a3  =   3.750000000000000e-01  #  3/8
    a4  =   9.230769230769231e-01  #  12/13
    a5  =   1.000000000000000e+00  #  1
    a6  =   5.000000000000000e-01  #  1/2

    # Coefficients used to compute the dependent variable argument of f
    b21 =   2.500000000000000e-01  #  1/4
    b31 =   9.375000000000000e-02  #  3/32
    b32 =   2.812500000000000e-01  #  9/32
    b41 =   8.793809740555303e-01  #  1932/2197
    b42 =  -3.277196176604461e+00  # -7200/2197
    b43 =   3.320892125625853e+00  #  7296/2197
    b51 =   2.032407407407407e+00  #  439/216
    b52 =  -8.000000000000000e+00  # -8
    b53 =   7.173489278752436e+00  #  3680/513
    b54 =  -2.058966861598441e-01  # -845/4104
    b61 =  -2.962962962962963e-01  # -8/27
    b62 =   2.000000000000000e+00  #  2
    b63 =  -1.381676413255361e+00  # -3544/2565
    b64 =   4.529727095516569e-01  #  1859/4104
    b65 =  -2.750000000000000e-01  # -11/40

    # Coefficients used to compute local truncation error estimate.  These
    # come from subtracting a 4th order RK estimate from a 5th order RK
    # estimate.
    r1  =   2.777777777777778e-03  #  1/360
    r3  =  -2.994152046783626e-02  # -128/4275
    r4  =  -2.919989367357789e-02  # -2197/75240
    r5  =   2.000000000000000e-02  #  1/50
    r6  =   3.636363636363636e-02  #  2/55

    # Coefficients used to compute 4th order RK estimate
    c1  =   1.157407407407407e-01  #  25/216
    c3  =   5.489278752436647e-01  #  1408/2565
    c4  =   5.353313840155945e-01  #  2197/4104
    c5  =  -2.000000000000000e-01  # -1/5

    # Set t and x according to initial condition and assume that h starts
    # with a value that is as large as possible.
    t = a
    x = x0
    h = hmax

    # Initialize arrays that will be returned
    T = np.array([t])
    X = np.array([x])

    while t < b:

        # Adjust step size when we get to last interval
        if t + h > b:
            h = b - t;

        # Compute values needed to compute truncation error estimate and
        # the 4th order RK estimate.
        F_thermal = np.sqrt(2*D/h)*np.random.normal(size=x.shape)
        k1 = h*(F_thermal + f(x, t))
        k2 = h*(F_thermal + f(x + b21*k1, t + a2*h))
        k3 = h*(F_thermal + f(x + b31*k1 + b32*k2, t + a3*h))
        k4 = h*(F_thermal + f(x + b41*k1 + b42*k2 + b43*k3, t + a4*h))
        k5 = h*(F_thermal + f(x + b51*k1 + b52*k2 + b53*k3 + b54*k4, t + a5*h))
        k6 = h*(F_thermal + f(x + b61*k1 + b62*k2 + b63*k3 + b64*k4 + b65*k5, t + a6*h))

        # Compute the estimate of the local truncation error.  If it's small
        # enough then we accept this step and save the 4th order estimate.
        r = abs( r1 * k1 + r3 * k3 + r4 * k4 + r5 * k5 + r6 * k6 ) / h
        if len( np.shape( r ) ) > 0:
            r = np.max( r )
        if r <= tol:
            t = t + h
            x = x + c1 * k1 + c3 * k3 + c4 * k4 + c5 * k5
            # x = x + np.sqrt(2*D*h)*np.random.normal(size=x.shape)
            T = np.append( T, t )
            X = np.append( X, [x], 0 )

        # # Now compute next step size, and make sure that it is not too big or
        # # too small.
        # h = h * min( max( 0.84 * ( tol / r )**0.25, 0.1 ), 4.0 )
        # if h > hmax:
        #     h = hmax
        # elif h < hmin:
        #     #TODO something's buggy with stepsize, see: rouse.rouse(101, 15,
        #     # 17475, 0, 0.0001, 16667), needs hmin=1e-25...
        #     print( "Error: minimum requested stepsize of %e exceeded." % hmin )
        #     break

    return T, X
